"""The :mod:`mpcrl.core` submodule implements various core functionalities of the MPC-RL
framework. These allow to define various hyperparameters to, e.g., specify the frequency
of the RL updates, use experience replay, introduce exploration in the MPC policy. At
the same time, it contains foundations for specifying the learnable parameters, and
implement the hook-callback mechanism that allows to customize the training loop to a
high degree."""
